@@ -3,7 +3,9 @@ from __future__ import absolute_import, division, print_function
 
 import functools
 import os
+import io
 from threading import Thread
+from datetime import timedelta
 
 from lostc import collection as lcc
 
@@ -20,16 +22,16 @@ from suanpan.utils import pbar as spbar
 
 class Storage(base.Storage):
     def __init__(
-        self,
-        minioAccessKey=None,
-        minioSecretKey=None,
-        minioBucketName="suanpan",
-        minioEndpoint="minio-serivce.default:9000",
-        minioSecure=True,
-        minioDelimiter="/",
-        minioTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
-        minioGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
-        **kwargs,
+            self,
+            minioAccessKey=None,
+            minioSecretKey=None,
+            minioBucketName="suanpan",
+            minioEndpoint="minio-serivce.default:9000",
+            minioSecure=True,
+            minioDelimiter="/",
+            minioTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
+            minioGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
+            **kwargs,
     ):
         super().__init__(
             delimiter=minioDelimiter,
@@ -46,11 +48,11 @@ class Storage(base.Storage):
     def _analyzeEndpoint(self, endpoint, secure=False):
         httpsPrefix = "https://"
         if endpoint.startswith(httpsPrefix):
-            return endpoint[len(httpsPrefix) :], True
+            return endpoint[len(httpsPrefix):], True
 
         httpPrefix = "http://"
         if endpoint.startswith(httpPrefix):
-            return endpoint[len(httpPrefix) :], False
+            return endpoint[len(httpPrefix):], False
 
         return endpoint, secure
 
@@ -87,13 +89,13 @@ class Storage(base.Storage):
 
     @runtime.retry(stop_max_attempt_number=3)
     def download(
-        self,
-        key,
-        path=None,
-        progress=None,
-        bucket=None,
-        ignores=None,
-        quiet=not g.debug,
+            self,
+            key,
+            path=None,
+            progress=None,
+            bucket=None,
+            ignores=None,
+            quiet=not g.debug,
     ):
         ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
         downloadFunction = (
@@ -106,14 +108,14 @@ class Storage(base.Storage):
         )
 
     def downloadFolder(
-        self,
-        key,
-        path=None,
-        progress=None,
-        bucket=None,
-        delimiter=None,
-        ignores=None,
-        quiet=not g.debug,
+            self,
+            key,
+            path=None,
+            progress=None,
+            bucket=None,
+            delimiter=None,
+            ignores=None,
+            quiet=not g.debug,
     ):
         bucket = bucket or self.bucket
         ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
@@ -126,7 +128,7 @@ class Storage(base.Storage):
 
         if self.basename(key) in ignores:
             with spbar.one(
-                total=1, desc="Downloading", disable=quiet, progress=progress
+                    total=1, desc="Downloading", disable=quiet, progress=progress
             ) as pbar:
                 pbar.update(1)
                 pbar.set_description("Ignored")
@@ -157,20 +159,22 @@ class Storage(base.Storage):
         asyncio.map(
             spath.remove, restFiles, pbar=quiet, desc="Removing Rest Files",
         )
-        spath.removeEmptyFolders(path)
+        # only if the `path` folder is getPathInTempStore, try to clear it
+        if path == self.getPathInTempStore(key):
+            spath.removeEmptyFolders(path)
         if not quiet:
             logger.debug(f"Removed empty folders in: {path}")
         # End
         return path
 
     def downloadFile(
-        self,
-        key,
-        path=None,
-        progress=None,
-        bucket=None,
-        ignores=None,
-        quiet=not g.debug,
+            self,
+            key,
+            path=None,
+            progress=None,
+            bucket=None,
+            ignores=None,
+            quiet=not g.debug,
     ):
         bucket = bucket or self.bucket
         ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
@@ -182,7 +186,7 @@ class Storage(base.Storage):
             logger.info(f"Downloading file: {url} -> {path}")
 
         with spbar.one(
-            total=fileSize, desc="Downloading", disable=quiet, progress=progress
+                total=fileSize, desc="Downloading", disable=quiet, progress=progress
         ) as pbar:
             if self.basename(key) in ignores:
                 pbar.update(fileSize)
@@ -206,13 +210,13 @@ class Storage(base.Storage):
 
     @runtime.retry(stop_max_attempt_number=3)
     def upload(
-        self,
-        name,
-        path=None,
-        progress=None,
-        bucket=None,
-        ignores=None,
-        quiet=not g.debug,
+            self,
+            name,
+            path=None,
+            progress=None,
+            bucket=None,
+            ignores=None,
+            quiet=not g.debug,
     ):
         bucket = bucket or self.bucket
         ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
@@ -222,13 +226,13 @@ class Storage(base.Storage):
         )
 
     def uploadFolder(
-        self,
-        key,
-        path=None,
-        progress=None,
-        bucket=None,
-        ignores=None,
-        quiet=not g.debug,
+            self,
+            key,
+            path=None,
+            progress=None,
+            bucket=None,
+            ignores=None,
+            quiet=not g.debug,
     ):
         bucket = bucket or self.bucket
         ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
@@ -240,7 +244,7 @@ class Storage(base.Storage):
 
         if self.basename(key) in ignores:
             with spbar.one(
-                total=1, desc="Uploading", disable=quiet, progress=progress
+                    total=1, desc="Uploading", disable=quiet, progress=progress
             ) as pbar:
                 pbar.update(1)
                 pbar.set_description("Ignored")
@@ -287,13 +291,13 @@ class Storage(base.Storage):
         return path
 
     def uploadFile(
-        self,
-        key,
-        path=None,
-        progress=None,
-        bucket=None,
-        ignores=None,
-        quiet=not g.debug,
+            self,
+            key,
+            path=None,
+            progress=None,
+            bucket=None,
+            ignores=None,
+            quiet=not g.debug,
     ):
         bucket = bucket or self.bucket
         ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
@@ -305,7 +309,7 @@ class Storage(base.Storage):
             logger.info(f"Uploading file: {path} -> {url}")
 
         with spbar.one(
-            total=fileSize, desc="Uploading", disable=quiet, progress=progress
+                total=fileSize, desc="Uploading", disable=quiet, progress=progress
         ) as pbar:
             if self.basename(key) in ignores:
                 pbar.update(fileSize)
@@ -339,7 +343,7 @@ class Storage(base.Storage):
         return copyFunction(src, dist, progress=progress, bucket=bucket, quiet=quiet)
 
     def copyFolder(
-        self, src, dist, progress=None, bucket=None, delimiter=None, quiet=not g.debug,
+            self, src, dist, progress=None, bucket=None, delimiter=None, quiet=not g.debug,
     ):
         bucket = bucket or self.bucket
         delimiter = delimiter or self.delimiter
@@ -372,7 +376,7 @@ class Storage(base.Storage):
             logger.info(f"Copying file: {srcurl} -> {disturl}")
 
         with spbar.one(
-            total=fileSize, desc="Copying", disable=quiet, progress=progress
+                total=fileSize, desc="Copying", disable=quiet, progress=progress
         ) as pbar:
             objectMd5 = self.getStorageMd5(src, bucket=bucket)
             distMd5 = self.getStorageMd5(dist, bucket=bucket)
@@ -389,7 +393,7 @@ class Storage(base.Storage):
 
     @runtime.retry(stop_max_attempt_number=3)
     def remove(
-        self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
+            self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
     ):
         delimiter = delimiter or self.delimiter
         bucket = bucket or self.bucket
@@ -399,7 +403,7 @@ class Storage(base.Storage):
         return removeFunc(key, delimiter=delimiter, bucket=bucket, quiet=quiet)
 
     def removeFolder(
-        self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
+            self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
     ):
         delimiter = delimiter or self.delimiter
         bucket = bucket or self.bucket
@@ -426,7 +430,7 @@ class Storage(base.Storage):
         return key
 
     def removeFile(
-        self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
+            self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
     ):  # pylint: disable=unused-argument
         delimiter = delimiter or self.delimiter
         bucket = bucket or self.bucket
@@ -436,7 +440,7 @@ class Storage(base.Storage):
 
         _remove = self.autoRefreshToken(self.client.remove_object)
         with spbar.one(
-            total=1, disable=quiet, desc="Removing", progress=progress,
+                total=1, disable=quiet, desc="Removing", progress=progress,
         ) as pbar:
             self.autoRefreshToken(self.client.remove_object)(bucket, key)
             pbar.update(1)
@@ -483,6 +487,45 @@ class Storage(base.Storage):
                 return False
             raise e
 
+    def presignedGetObject(self, key, expires=3600, bucket=None):
+        bucket = bucket or self.bucket
+        url = self.client.presigned_get_object(bucket, key, expires=timedelta(seconds=expires))
+        return url
+
+    def presignedPutObject(self, key, expires=3600, bucket=None):
+        bucket = bucket or self.bucket
+        url = self.client.presigned_put_object(bucket, key, expires=timedelta(seconds=expires))
+        return url
+
+    def getObject(self, key, bucket=None, **kwargs):
+        logger.debug(f"Get object from {key}")
+        bucket = bucket or self.bucket
+        response = None
+        try:
+            response = self.autoRefreshToken(self.client.get_object)(bucket, key, **kwargs)
+            return response.read()
+        finally:
+            if response:
+                response.close()
+                response.release_conn()
+
+    def putObject(self, key, data, bucket=None, **kwargs):
+        logger.debug(f"Put object to {key}")
+        bucket = bucket or self.bucket
+        length = kwargs.get('length')
+        part_size = kwargs.get('part_size')
+        if not length and not part_size:
+            length = -1
+            part_size = 10 * 1024 * 1024
+
+        if isinstance(data, str):
+            data = io.BytesIO(data.encode("utf8"))
+        elif isinstance(data, bytes):
+            data = io.BytesIO(data)
+        result = self.autoRefreshToken(self.client.put_object)(
+            bucket, key, data, length=length, part_size=part_size, **kwargs)
+        return result
+
     def getStorageMd5(self, key, bucket=None):
         bucket = bucket or self.bucket
         try:
